#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import absolute_import

import collections

from hexdi import lifetime, gentype as __gentype__
from hexdi import loader
from hexdi import finder
from hexdi.core import get_root_container, clstype as __clstype__
# Quick access imports
from hexdi.decorator import component, permanent, transient, dependency, inject


# Shortcut functions

def resolve(accessor: __clstype__) -> __gentype__.T:
    """
    shortcut for resolving from root container

    :param accessor: accessor for resolving object
    :return: resolved object of requested type
    """
    return get_root_container().resolve(accessor=accessor)


def get_loader(modules: collections.Iterable) -> loader.AbstractBaseLoader:
    """
    shortcut for constructing of basic loader

    :param modules: list of modules to load
    :return: constructed basic class loader
    """
    return loader.BasicLoader(modules)


def get_finder(packages: collections.Iterable) -> finder.AbstractBaseFinder:
    """
    shortcut for constructing of recursive regex finder

    :param packages:
    """
    return finder.RecursiveRegexFinder(packages)



#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import typing
import hexdi
import hexdi.gentype
import hexdi.lifetime


def component(lifetime, accessor=None):
    def component_wrap(type_to_bind):
        container = hexdi.get_root_container()
        if accessor is None:
            container.bind_type(accessor=type_to_bind, type_to_resolve=type_to_bind, lifetime_manager=lifetime)
        else:
            container.bind_type(accessor=accessor, type_to_resolve=type_to_bind, lifetime_manager=lifetime)
        return type_to_bind

    return component_wrap


def permanent(accessor=None):
    return component(hexdi.lifetime.PermanentLifeTimeManager, accessor)


def transient(accessor=None):
    return component(hexdi.lifetime.PerResolveLifeTimeManager, accessor)


def dependency(cls: typing.Type[hexdi.gentype.T]):
    def dependency_func(func):
        def dependency_func_injection(*args, **kwargs) -> hexdi.gentype.T:
            container = hexdi.get_root_container()
            instance = container.resolve(cls)
            return instance

        return dependency_func_injection

    return dependency_func


def inject(*deps: typing.Optional[typing.Type[hexdi.gentype.T]]):
    def inject_func(func):
        def inject_func_args(*args):
            container = hexdi.get_root_container()
            resolves = list(args)
            for d in deps:
                resolves.append(container.resolve(d))
            return func(*resolves)

        return inject_func_args

    return inject_func



#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import collections
import importlib
import os
import site


def __get_root_package_directories():
    current_working_dir = os.getcwd()
    root_packages = site.getsitepackages()
    return root_packages + [current_working_dir]


__DIRECTORIES_TO_SEARCH_PACKAGES = __get_root_package_directories()


def get_module_name(obj):
    if isinstance(obj, str):
        return obj

    if callable(obj):
        return get_module_name(obj.__call__())

    if isinstance(obj, collections.Iterable):
        lst = list(obj)
        if len(lst) > 0:
            o = lst.pop(0)
            if callable(o):
                if len(lst) > 0:
                    arguments = lst[1]
                    return get_module_name(o.__call__(*arguments))


def __sanitize_path(mod: str):
    path = mod.replace('.', '/')
    return list([os.path.join(d, path) for d in __DIRECTORIES_TO_SEARCH_PACKAGES])


def __sanitize_module(mod: str):
    paths = __sanitize_path(mod)
    return list(["{}.py".format(path) for path in paths])


def get_submodules(mod):
    package = get_package_if_exists(mod)
    if package is not None:
        lst = os.listdir(package)
        submodules = set()
        for l in lst:
            if l in ['__pycache__']:
                continue
            if __is_module_without_sanitizing(package, l) and l.endswith('.py'):
                submodules.add(l[:-3])
            else:
                submodules.add(l)
        return list(["{}.{}".format(mod, l) for l in submodules])
    else:
        return []


def get_module_if_exists(mod):
    mods = __sanitize_module(mod)
    for m in mods:
        if os.path.isfile(m):
            return m
    return None


def get_package_if_exists(mod):
    paths = __sanitize_path(mod)
    for p in paths:
        if os.path.isdir(p):
            return p
    return None


def __is_module_without_sanitizing(path, mod):
    return os.path.isfile(os.path.join(path, mod))


def read_module_as_string(mod):
    mod_path = get_module_if_exists(mod)
    with open(mod_path, encoding='utf-8', mode='r') as f:
        return f.read()


def load_module(mod: str):
    return importlib.import_module(mod, package=__name__)


def load_class(path: str):
    mod, cls = path.rsplit('.')
    module_path = ".".join(mod)
    _module = load_module(module_path)
    return getattr(_module, cls)


def loading_indicator(percentage, prefix='', suffix='', indicator='❚'):
    percentage_int = int(percentage)
    if 0 <= percentage_int <= 100:
        free_space_int = 100 - percentage_int
        filled = indicator * percentage_int
        free = '-' * free_space_int
        print('%s [%s%s] %s%% %s' % (prefix, filled, free, percentage_int, suffix), end='\r')



#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import importlib
import typing
from hexdi import utils


class AbstractBaseLoader(object):
    def load(self):
        pass


class AbstractOneTimeLoader(AbstractBaseLoader):
    def __init__(self):
        self.__loaded = False

    def load(self):
        """
        load classes to providing auto register
        """
        if not self.__loaded:
            self._do_loading()
            self.__loaded = True

    def _do_loading(self):
        pass


class BasicLoader(AbstractOneTimeLoader):
    def __init__(self, modules: typing.Iterable):
        super().__init__()
        self.__modules = modules

    def _do_loading(self):
        for mod in self.__modules:
            utils.load_module(utils.get_module_name(mod))



#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import typing
import re
from hexdi import decorator
from hexdi import core
from hexdi import utils


class AbstractBaseFinder(object):
    def find(self) -> typing.List[typing.AnyStr]:
        pass


class AbstractOneTimeFinder(AbstractBaseFinder):
    def __init__(self):
        self.__found = False
        self.__modules = None

    def _do_find(self) -> typing.List[typing.AnyStr]:
        pass

    def find(self) -> typing.List[typing.AnyStr]:
        if not self.__found:
            self.__modules = self._do_find()
            self.__found = True
        return self.__modules


class RecursiveRegexFinder(AbstractOneTimeFinder):
    __component_re = re.compile(r'@?(?:hexdi\.)?{}'.format(decorator.component.__name__), re.MULTILINE)
    __permanent_re = re.compile(r'@?(?:hexdi\.)?{}'.format(decorator.permanent.__name__), re.MULTILINE)
    __transient_re = re.compile(r'@?(?:hexdi\.)?{}'.format(decorator.transient.__name__), re.MULTILINE)
    __bind_type_re = re.compile(r'[^.]{}'.format(core.DiContainer.bind_type.__name__), re.MULTILINE)
    __bind_inst_re = re.compile(r'[^.]{}'.format(core.DiContainer.bind_instance.__name__), re.MULTILINE)

    __regex_list = [
        __component_re,
        __permanent_re,
        __transient_re,
        __bind_type_re,
        __bind_inst_re
    ]

    def __init__(self, packages):
        super().__init__()
        self.__packages = packages
        self.__found_modules = set()
        self.__context = set()

    def __parse(self, mod):
        module_string = utils.read_module_as_string(mod)
        for regex in self.__regex_list:
            if regex.search(module_string) is not None:
                self.__context.add(mod)
                break

    def __investigate(self, mod):
        if utils.get_package_if_exists(mod):
            submodules = utils.get_submodules(mod)
            for m in submodules:
                self.__investigate(m)
        elif utils.get_module_if_exists(mod):
            self.__found_modules.add(mod)

    def _do_find(self) -> typing.List[typing.AnyStr]:
        for mod in self.__packages:
            self.__investigate(mod)
        for mod in self.__found_modules:
            self.__parse(mod)
        return list(self.__context)



#    Highly extensible Dependency injection framework for humans
#    Copyright (C) 2017 Dmitriy Selischev
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import collections
import typing
from hexdi import gentype
from hexdi import meta
from hexdi import lifetime
from hexdi import utils

clstype = typing.Union[str, typing.Type[gentype.T]]
restype = typing.Union[str, gentype.T]
ltype_instance = lifetime.BaseLifeTimeManager
ltype = typing.Type[ltype_instance]


class DiContainer(metaclass=meta.DiContainerMeta):
    def __init__(self):
        self.container = dict()

    @staticmethod
    def _sanitize_accessor(o):
        if isinstance(o, str):
            return o
        else:
            return o.__name__

    def binded(self, accessor) -> bool:
        return self.container.get(self._sanitize_accessor(accessor)) is not None

    def bind_type(self, type_to_resolve: restype, accessor, lifetime_manager: ltype) -> None:
        if isinstance(type_to_resolve, str):
            type_to_resolve = utils.load_class(type_to_resolve)
        if isinstance(accessor, str) or issubclass(type_to_resolve, accessor):
            self._bind_type(accessor, lifetime_manager(type_to_resolve))
        elif isinstance(accessor, collections.Iterable):
            for element in accessor:
                self.bind_type(type_to_resolve, element, lifetime_manager)

    def _bind_type(self, accessor, lifetime_manager: ltype_instance):
        if not self.binded(accessor):
            self.container[self._sanitize_accessor(accessor)] = lifetime_manager
        else:
            raise Exception('already registered instance by \'%s\'' % self._sanitize_accessor(accessor))

    def bind_instance(self, instance: gentype.T, accessor):
        self._bind_type(accessor, lifetime.PreparedInstanceLifeTimeManager(instance))

    def unbind(self, accessor) -> None:
        if self.binded(accessor):
            del self.container[self._sanitize_accessor(accessor)]
        else:
            raise Exception('class %s is not registered' % self._sanitize_accessor(accessor))

    def resolve(self, accessor: clstype) -> gentype.T:
        if self.binded(accessor):
            return self.container.get(self._sanitize_accessor(accessor)).resolve()
        else:
            raise Exception('class %s is not registered' % self._sanitize_accessor(accessor))

    def destroy(self):
        self.container.clear()


def get_root_container() -> DiContainer:
    return meta.get_root_container()
